@@ -11,35 +11,40 @@ import (
 	"github.com/AdguardTeam/golibs/log"
 )
 
+// dialTimeout is the global timeout for establishing a TLS connection.
+// TODO(ameshkov): use bootstrap timeout instead.
 const dialTimeout = 10 * time.Second
 
 // TLSPool is a connections pool for the DNS-over-TLS Upstream.
 //
 // Example:
-//  pool := TLSPool{Address: "tls://1.1.1.1:853"}
-//  netConn, err := pool.Get()
-//  if err != nil {panic(err)}
-//  c := dns.Conn{Conn: netConn}
-//  q := dns.Msg{}
-//  q.SetQuestion("google.com.", dns.TypeA)
-//  log.Println(q)
-//  err = c.WriteMsg(&q)
-//  if err != nil {panic(err)}
-//  r, err := c.ReadMsg()
-//  if err != nil {panic(err)}
-//  log.Println(r)
-//  pool.Put(c.Conn)
+//
+//	pool := TLSPool{Address: "tls://1.1.1.1:853"}
+//	netConn, err := pool.Get()
+//	if err != nil {panic(err)}
+//	c := dns.Conn{Conn: netConn}
+//	q := dns.Msg{}
+//	q.SetQuestion("google.com.", dns.TypeA)
+//	log.Println(q)
+//	err = c.WriteMsg(&q)
+//	if err != nil {panic(err)}
+//	r, err := c.ReadMsg()
+//	if err != nil {panic(err)}
+//	log.Println(r)
+//	pool.Put(c.Conn)
 type TLSPool struct {
 	boot *bootstrapper
 
-	// connections
-	conns      []net.Conn
-	connsMutex sync.Mutex // protects conns
+	// conns is the list of connections available in the pool.
+	conns []net.Conn
+	// connsMutex protects conns.
+	connsMutex sync.Mutex
 }
 
-// Get gets or creates a new TLS connection
+// Get gets a connection from the pool (if there's one available) or creates
+// a new TLS connection.
 func (n *TLSPool) Get() (net.Conn, error) {
-	// get the connection from the slice inside the lock
+	// Get the connection from the slice inside the lock.
 	var c net.Conn
 	n.connsMutex.Lock()
 	num := len(n.conns)
@@ -50,7 +55,7 @@ func (n *TLSPool) Get() (net.Conn, error) {
 	}
 	n.connsMutex.Unlock()
 
-	// if we got connection from the slice, update deadline and return it.
+	// If we got connection from the slice, update deadline and return it.
 	if c != nil {
 		err := c.SetDeadline(time.Now().Add(dialTimeout))
 
@@ -64,7 +69,7 @@ func (n *TLSPool) Get() (net.Conn, error) {
 	return n.Create()
 }
 
-// Create creates a new connection for the pool (but not puts it there)
+// Create creates a new connection for the pool (but not puts it there).
 func (n *TLSPool) Create() (net.Conn, error) {
 	tlsConfig, dialContext, err := n.boot.get()
 	if err != nil {
@@ -80,7 +85,7 @@ func (n *TLSPool) Create() (net.Conn, error) {
 	return conn, nil
 }
 
-// Put returns connection to the pool
+// Put returns the connection to the pool.
 func (n *TLSPool) Put(c net.Conn) {
 	if c == nil {
 		return
@@ -90,16 +95,18 @@ func (n *TLSPool) Put(c net.Conn) {
 	n.connsMutex.Unlock()
 }
 
-// tlsDial is basically the same as tls.DialWithDialer, but we will call our own dialContext function to get connection
+// tlsDial is basically the same as tls.DialWithDialer, but we will call our own
+// dialContext function to get connection.
 func tlsDial(dialContext dialHandler, network string, config *tls.Config) (*tls.Conn, error) {
-	// we're using bootstrapped address instead of what's passed to the function
+	// We're using bootstrapped address instead of what's passed
+	// to the function.
 	rawConn, err := dialContext(context.Background(), network, "")
 	if err != nil {
 		return nil, err
 	}
 
-	// we want the timeout to cover the whole process: TCP connection and TLS handshake
-	// dialTimeout will be used as connection deadLine
+	// We want the timeout to cover the whole process: TCP connection and
+	// TLS handshake dialTimeout will be used as connection deadLine.
 	conn := tls.Client(rawConn, config)
 	err = conn.SetDeadline(time.Now().Add(dialTimeout))
 	if err != nil {
